"""
This modules contains classifiers which classify a given coarse grained
geometry as a certain type of interaction.

An example would be classifying relative element orientations as A-minor
interaction or stacking interaction.

Classifiers from this module are compatible with scikit-learn.
"""
